
/* demo_key.c: Sony Demo Key Driver.
 *
 * Copyright 2016 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <asm/uaccess.h>
#include <mach/mt_gpio.h>
#include <asm/atomic.h>
#include <mach/irqs.h>
#include <mach/eint.h>
#include <cust_eint.h>
#include <mt_auxadc.h>
#include <icx_key.h>
#include "./demo_key.h"

#define _DEBUG_DEMO_KEY_
#ifdef _DEBUG_DEMO_KEY_
#define TRACE(fmt, ...)	 	printk("[KEY][L.%d] %s: " fmt , __LINE__, __func__, ## __VA_ARGS__)
#define ERROR(fmt, ...)	 	printk("[KEY][L.%d] ### ERROR %s: " fmt , __LINE__, __func__, ## __VA_ARGS__)
#else /* _DEBUG_DEMO_KEY_ */
#define TRACE(fmt, ...)	 	do {} while (0)
#define ERROR(fmt, ...)	 	printk("[KEY] ### ERROR %s: " fmt , __func__, ## __VA_ARGS__)
#endif /* _DEBUG_DEMO_KEY_ */

#define DRV_NAME 	 	 	 	"demo_key"

#define AUX_IN2		 	 	2		/* one channels: AUX_IN2 */
static u8 adc_ch_key2 = AUX_IN2;

static u8 demo_status;
#define DEMO_STATUS_MODEOUT	0			/* demo mode out */
#define	DEMO_STATUS_MODEIN	1			/* demo mode in */
#define	DEMO_STATUS_KEYPUSH	2			/* demo mode in */

static int demo_key_code     = KEY_RESERVED;
static int pre_demo_key_code = KEY_RESERVED;

#define POLLING_FOR_MODE_IN		 	500	// msec
#define NORMAL_POLL_INTERVAL	 	10	// msec
#define SCAN_INTERVAL_CHAT_GUARD	10	// msec
#define ADIN2_ERROR_TRIAL_WAIT		50	// msec
#define CHAT_TRIAL_CNT				3
#define ADIN2_ERROR_TRIAL_CNT		3

/* AD KEY */
/* AD2 Center */
#define KC_AD2_0		 	 	KEY_DEMO_IN
#define KC_AD2_1		 	 	KEY_DEMO_AUDIO
#define KC_AD2_2		 	 	KEY_DEMO_BT
#define KC_AD2_3		 	 	KEY_DEMO_VOL_P
#define KC_AD2_4		 	 	KEY_DEMO_VOL_M
#define KC_AD2_5		 	 	KEY_DEMO_USB_PLAY
#define KC_AD2_6		 	 	KEY_DEMO_USB_FF
#define KC_AD2_7		 	 	KEY_DEMO_USB_LOAD
#define KC_AD2_8		 	 	KEY_DEMO_VOL_RST

/* AD2 MIN */
#define KC_AD2_0_MIN	 	 	KEY_DEMO_IN_MIN
#define KC_AD2_1_MIN	 	 	KEY_DEMO_AUDIO_MIN
#define KC_AD2_2_MIN	 	 	KEY_DEMO_BT_MIN
#define KC_AD2_3_MIN	 	 	KEY_DEMO_VOL_P_MIN
#define KC_AD2_4_MIN	 	 	KEY_DEMO_VOL_M_MIN
#define KC_AD2_5_MIN	 	 	KEY_DEMO_USB_PLAY_MIN
#define KC_AD2_6_MIN	 	 	KEY_DEMO_USB_FF_MIN
#define KC_AD2_7_MIN	 	 	KEY_DEMO_USB_LOAD_MIN
#define KC_AD2_8_MIN	 	 	KEY_DEMO_VOL_RST_MIN

/* AD2 MAX */
#define KC_AD2_0_MAX	 	 	KEY_DEMO_IN_MAX
#define KC_AD2_1_MAX	 	 	KEY_DEMO_AUDIO_MAX
#define KC_AD2_2_MAX	 	 	KEY_DEMO_BT_MAX
#define KC_AD2_3_MAX	 	 	KEY_DEMO_VOL_P_MAX
#define KC_AD2_4_MAX	 	 	KEY_DEMO_VOL_M_MAX
#define KC_AD2_5_MAX	 	 	KEY_DEMO_USB_PLAY_MAX
#define KC_AD2_6_MAX	 	 	KEY_DEMO_USB_FF_MAX
#define KC_AD2_7_MAX	 	 	KEY_DEMO_USB_LOAD_MAX
#define KC_AD2_8_MAX	 	 	KEY_DEMO_VOL_RST_MAX

#define INPUT_EVENT_UP 	 		DEMO_KEY_VALUE_UP
#define INPUT_EVENT_DOWN		DEMO_KEY_VALUE_DOWN

#define	DEMO_KEY_COUNT				9
#define	DEMO_KEY_CODE_DEMO_IN_OUT	KEY_1
#define	DEMO_KEY_CODE_BT			KEY_2
#define	DEMO_KEY_CODE_AUDIO			KEY_3
#define	DEMO_KEY_CODE_VOL_P			KEY_4
#define	DEMO_KEY_CODE_VOL_M			KEY_5
#define	DEMO_KEY_CODE_USB_P			KEY_6
#define	DEMO_KEY_CODE_USB_FF		KEY_7
#define	DEMO_KEY_CODE_USB_L			KEY_8
#define	DEMO_KEY_CODE_VOL_R			KEY_9

typedef struct {
	struct miscdevice		misc;
	struct input_dev		*p_input;
	struct delayed_work		dwork;
	u8						key_count;
	u32						p_key_code[DEMO_KEY_COUNT];
} demo_key_dev_t;

static demo_key_dev_t				*m_p_demo_key;

typedef struct {
	int key_ad_min;
	int key_ad_max;
	int key_code;
} adc_key_range_t;

static const adc_key_range_t ad_2_code_table[] = {
	// for DemoMode AD key range specification
	{KC_AD2_0_MIN, KC_AD2_0_MAX, DEMO_KEY_CODE_DEMO_IN_OUT},
	{KC_AD2_1_MIN, KC_AD2_1_MAX, DEMO_KEY_CODE_AUDIO},
	{KC_AD2_2_MIN, KC_AD2_2_MAX, DEMO_KEY_CODE_BT},
	{KC_AD2_3_MIN, KC_AD2_3_MAX, DEMO_KEY_CODE_VOL_P},
	{KC_AD2_4_MIN, KC_AD2_4_MAX, DEMO_KEY_CODE_VOL_M},
	{KC_AD2_5_MIN, KC_AD2_5_MAX, DEMO_KEY_CODE_USB_P},
	{KC_AD2_6_MIN, KC_AD2_6_MAX, DEMO_KEY_CODE_USB_FF},
	{KC_AD2_7_MIN, KC_AD2_7_MAX, DEMO_KEY_CODE_USB_L},
	{KC_AD2_8_MIN, KC_AD2_8_MAX, DEMO_KEY_CODE_VOL_R},
};

static void key_sleep(int timeout)
{
  set_current_state(TASK_INTERRUPTIBLE);
  schedule_timeout(timeout);
}

static void input_sync_key_down(int key_code)
{
	input_event(m_p_demo_key->p_input, EV_KEY, key_code, DEMO_KEY_VALUE_DOWN);
	input_sync(m_p_demo_key->p_input);
}

static void input_sync_key_up(int key_code)
{
	input_event(m_p_demo_key->p_input, EV_KEY, key_code, DEMO_KEY_VALUE_UP);
	input_sync(m_p_demo_key->p_input);
}

static inline int key_code_from_ad_val(int ad_val)
{
	int i;
	for (i = 0; i < ARRAY_SIZE(ad_2_code_table); i++)
		if ((ad_2_code_table[i].key_ad_min <= ad_val) && 
			(ad_val <= ad_2_code_table[i].key_ad_max)) {
			return ad_2_code_table[i].key_code;
		}
	return -1;
}

static int get_md_adc_val(unsigned int num)
{
	int data[4] = { 0, 0, 0, 0 };
	int val = 0;
	int pre_val = 0;
	int ret = 0;
	
	int key_code_pre = KEY_RESERVED;
	int key_code_cur = KEY_RESERVED;
	int chat_cnt     = 0;
	
	while (chat_cnt < CHAT_TRIAL_CNT) {
		if (chat_cnt) {
			key_sleep(msecs_to_jiffies(SCAN_INTERVAL_CHAT_GUARD));
		}
		ret = IMM_GetOneChannelValue(num, data, &val);
		//TRACE("adc val=%d\n", val>>4);
		if (ret == 0) {
			val = (val>>4);
			key_code_cur = key_code_from_ad_val(val);
			if ( (key_code_pre == KEY_RESERVED) || (key_code_pre == key_code_cur) ) {
				key_code_pre = key_code_cur;
				chat_cnt++;
			} else {
				key_code_pre = KEY_RESERVED;
				key_code_cur = KEY_RESERVED;
				chat_cnt     = 0;
			}
		} else {
			return -1;
		}		
	}
	return val;
}

static void workq_func_demo(struct work_struct *work)
{
	int ad_val;
	int i = 0;

	// reading ADIN2 value
	for(i=0; i<ADIN2_ERROR_TRIAL_CNT; i++) {
		ad_val = get_md_adc_val(adc_ch_key2);
		if(!(ad_val < 0)) {
			break;
		}
		key_sleep(msecs_to_jiffies(ADIN2_ERROR_TRIAL_WAIT));
	}
	if (ad_val < 0) {
		ERROR("get_md_adc_val(ADIN%d) failed\n", adc_ch_key2 + 1);
		return;
	}    
	demo_key_code = key_code_from_ad_val(ad_val);

	switch (demo_status) {
	case DEMO_STATUS_MODEOUT:
		if (demo_key_code == DEMO_KEY_CODE_DEMO_IN_OUT) {
			/* demo jig plugged */
			input_sync_key_down(DEMO_KEY_CODE_DEMO_IN_OUT);
			demo_status = DEMO_STATUS_MODEIN;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( NORMAL_POLL_INTERVAL ));
			TRACE("[Demo Driver] Demo In\n");
		} else {
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( POLLING_FOR_MODE_IN ));
		}
		break;
	case DEMO_STATUS_MODEIN:
		if (ad_val > KEY_DEMO_IN_MAX) {
			/* demo jig unplugged */
			input_sync_key_up(DEMO_KEY_CODE_DEMO_IN_OUT);
			demo_status = DEMO_STATUS_MODEOUT;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( POLLING_FOR_MODE_IN ));
			TRACE("[Demo Driver] Demo Out\n");
		} else {
			if ((pre_demo_key_code != demo_key_code) && 
				((DEMO_KEY_CODE_DEMO_IN_OUT <= demo_key_code) && (demo_key_code <= DEMO_KEY_CODE_VOL_R))) {
				pre_demo_key_code = demo_key_code;
				if (demo_key_code != DEMO_KEY_CODE_DEMO_IN_OUT) {
					/* demo key pushed */
					input_sync_key_down(pre_demo_key_code);
					demo_status = DEMO_STATUS_KEYPUSH;
					TRACE("key code=%d\n", pre_demo_key_code);
					TRACE("key pressed\n");
				}
			}
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( NORMAL_POLL_INTERVAL ));
		}
		break;
	case DEMO_STATUS_KEYPUSH:
		/* polling for demo key release */
		input_sync_key_up(pre_demo_key_code);
		demo_status = DEMO_STATUS_MODEIN;
		TRACE("key code=%d\n", pre_demo_key_code);
		TRACE("key released\n");
		
		if (ad_val > KEY_DEMO_IN_MAX) {
			/* demo jig unplugged */
			input_sync_key_up(DEMO_KEY_CODE_DEMO_IN_OUT);
			demo_status = DEMO_STATUS_MODEOUT;
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( POLLING_FOR_MODE_IN ));
			TRACE("[Demo Driver] Demo Out\n");
		} else {
			schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( NORMAL_POLL_INTERVAL ));
		}
		break;
	default:
		break;
	}
}

static int dm_key_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data, demo_key_dev_t, misc) ;

	return 0 ;
}

static long dm_key_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	demo_key_dev_t	*p_demo_key=p_file->private_data;

	switch( cmd ) {
		case ICX_KEY_IOCTL_CMD_GETCOUNT:
		{
			return p_demo_key->key_count ;
		}

		case ICX_KEY_IOCTL_CMD_GETINFO:
		{
			if( copy_to_user((void __user *)arg, &p_demo_key->p_key_code[0], sizeof( u32 )*p_demo_key->key_count) ) {
				return -EFAULT ;
			}
			break ;
		}

		default:
			return -EINVAL ;
	}
	return 0 ;
}

static const struct file_operations		dm_key_fops={
											.owner          = THIS_MODULE,
											.open           = dm_key_open,
											.unlocked_ioctl = dm_key_ioctl,
											.compat_ioctl   = dm_key_ioctl,
										} ;

static int dm_key_suspend(struct platform_device *p_dev, pm_message_t state)
{
	demo_key_dev_t	*p_demo_key = platform_get_drvdata(p_dev);
	TRACE("entering suspend\n");
	cancel_delayed_work( &p_demo_key->dwork ) ;
	
	/* demo jig unplugged */
	input_sync_key_up(DEMO_KEY_CODE_DEMO_IN_OUT);
	demo_status = DEMO_STATUS_MODEOUT;
	schedule_delayed_work(&m_p_demo_key->dwork, msecs_to_jiffies( POLLING_FOR_MODE_IN ));
	TRACE("[Demo Driver] Demo Out\n");
	
	return 0;
}

static int dm_key_resume( struct platform_device *p_dev )
{
	demo_key_dev_t	*p_demo_key = platform_get_drvdata(p_dev);
	TRACE("entering resume\n");
	schedule_delayed_work(&p_demo_key->dwork, 0);
	return 0;
}

static int dm_key_remove( struct platform_device *p_dev )
{
	demo_key_dev_t	*p_demo_key = platform_get_drvdata(p_dev);
	TRACE("entering remove\n");
	cancel_delayed_work_sync( &p_demo_key->dwork ) ;
	input_unregister_device(p_demo_key->p_input);
	misc_deregister(&p_demo_key->misc);
	kfree(p_demo_key);
	return 0;
}

static int dm_key_probe( struct platform_device *p_dev )
{
	int				result;
	demo_key_dev_t	*p_demo_key;
	uint			count, size, i;
	demo_status = DEMO_STATUS_MODEOUT;
	count = DEMO_KEY_COUNT;
	
	printk(KERN_INFO "Starting Demo Mode Key Driver.\n");

	size = sizeof(demo_key_dev_t);
	p_demo_key = kzalloc(size, GFP_KERNEL);
	if( p_demo_key == NULL ) {
		printk("******** error(%d)\n", __LINE__) ;
		return -ENOMEM;
	}
	p_demo_key->key_count = count;
	INIT_DELAYED_WORK(&p_demo_key->dwork, workq_func_demo);

	/* MISC device register */
	p_demo_key->misc.minor = MISC_DYNAMIC_MINOR ;
	p_demo_key->misc.name  = DRV_NAME ;
	p_demo_key->misc.fops  = &dm_key_fops ;
	result = misc_register(&p_demo_key->misc) ;
	if( result ) {
		printk("******** error(%d)\n", __LINE__) ;
		kfree( p_demo_key ) ;
		return result;
	}

	/* allocate input_dev and initialize it. */
	p_demo_key->p_input = input_allocate_device();
	if( p_demo_key->p_input == NULL ) {
		result = -ENOMEM ;
		printk("******** error(%d)\n", __LINE__) ;
		misc_deregister( &p_demo_key->misc ) ;
		return result;
	}

	set_bit(EV_KEY, p_demo_key->p_input->evbit) ;
	p_demo_key->p_input->name       = DRV_NAME ;
	p_demo_key->p_input->phys       = DRV_NAME ;
	p_demo_key->p_input->id.bustype = BUS_HOST ;
	p_demo_key->p_input->id.vendor  = 0x0001 ;
	p_demo_key->p_input->id.product = 0x0001 ;
	p_demo_key->p_input->id.version = 0x0100 ;
	p_demo_key->p_input->dev.parent = &p_dev->dev ;
	
	p_demo_key->p_key_code[0] = DEMO_KEY_CODE_DEMO_IN_OUT;
	set_bit(DEMO_KEY_CODE_DEMO_IN_OUT, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[1] = DEMO_KEY_CODE_BT;
	set_bit(DEMO_KEY_CODE_BT, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[2] = DEMO_KEY_CODE_AUDIO;
	set_bit(DEMO_KEY_CODE_AUDIO, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[3] = DEMO_KEY_CODE_VOL_P;
	set_bit(DEMO_KEY_CODE_VOL_P, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[4] = DEMO_KEY_CODE_VOL_M;
	set_bit(DEMO_KEY_CODE_VOL_M, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[5] = DEMO_KEY_CODE_USB_P;
	set_bit(DEMO_KEY_CODE_USB_P, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[6] = DEMO_KEY_CODE_USB_FF;
	set_bit(DEMO_KEY_CODE_USB_FF, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[7] = DEMO_KEY_CODE_USB_L;
	set_bit(DEMO_KEY_CODE_USB_L, p_demo_key->p_input->keybit);
	p_demo_key->p_key_code[8] = DEMO_KEY_CODE_VOL_R;
	set_bit(DEMO_KEY_CODE_VOL_R, p_demo_key->p_input->keybit);

	for (i=0; i<REP_CNT; i++) {
		p_demo_key->p_input->rep[i] = 0;
	}

	if ((result = input_register_device(p_demo_key->p_input)) != 0) {
		ERROR("input_register_device() failed: result=%d\n", result);
		input_free_device( p_demo_key->p_input ) ;
	}

	platform_set_drvdata(p_dev, p_demo_key) ;
	m_p_demo_key = p_demo_key;

	/* going to scan ad_key right away. */
	schedule_delayed_work(&p_demo_key->dwork, 0);
	return 0 ;
}

static struct platform_driver		dm_key_platform_driver={
										.probe     = dm_key_probe,
										.remove    = dm_key_remove,
										.suspend   = dm_key_suspend,
										.resume    = dm_key_resume,
										.driver    = {
											.name  = DRV_NAME,
											.owner = THIS_MODULE,
										},
									} ;


module_platform_driver( dm_key_platform_driver ) ;
MODULE_DESCRIPTION("Sony Demo Key Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL") ;
MODULE_ALIAS("Sony Demo Key Driver") ;
MODULE_VERSION( "v0.01" ) ;

