/*
 *
 * Copyright (C) 2011 Sony Corporation. All rights reserved.
 *
 */

#include <fcntl.h>
#include <unistd.h>
#include <sys/select.h> // fuji
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/input.h>
#include <stdio.h>

#include "../demo_key.h"

static const struct {
  const __u16 code;
  const char *name;
} key_info[] = {
  /* AD2 */
  { 10		, "DEMO MODE IN[Key_9]" },
  { 3		, "AUDIO[Key_2]" },
  { 2		, "BT[Key_1]" },
  { 4		, "VOL +[Key_3]" },
  { 5		, "VOL -[Key_4]" },
  { 6		, "USB PLAY[Key_5]" },
  { 7		, "USB FF[Key_6]" },
  { 8		, "USB LOAD[Key_7]" },
  { 9		, "VOL RESET[Key_8]" },
};

static const char *key_name_from_code(int code)
{
  int i;
  for (i = 0; i < (int)sizeof(key_info) / sizeof(key_info[0]); i++) {
    if (code == key_info[i].code) {
      return key_info[i].name;
    }
  }
  return 0;
}

int main (int argc, char *argv[])
{
  int ret;
  int fd;
  fd_set rfds;
  struct input_event event;
  //const char *input_device = "/dev/input/event4";
  const char *input_device = "/dev/input/event6";
//-------------

//-------------  
  const char *name;

  if ((fd = open(input_device, O_RDONLY)) == -1) {
    printf("Can not open %s \n", input_device);
    return -1;
  }
  else {
      printf("Success to open  %s \n", input_device);
  }
	
  while (1) {
  
    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);
    
    ret = select(fd + 1, &rfds, NULL, NULL, NULL);
    if (ret == -1) {
      perror("select()");
    }
	
    read(fd, &event, sizeof(struct input_event));

{
	unsigned char key_b[KEY_MAX/8 + 1];
	memset(key_b, 0, sizeof(key_b));
	ioctl(fd, EVIOCGKEY(sizeof(key_b)), key_b);
	printf("[demo in] %d\n", !!(key_b[KEY_9/8] & (1<<(KEY_9 % 8))));
	printf("[audio] %d\n", !!(key_b[KEY_2/8] & (1<<(KEY_2 % 8))));
	printf("[BT] %d\n", !!(key_b[KEY_1/8] & (1<<(KEY_1 % 8))));
	printf("[VOL+] %d\n", !!(key_b[KEY_3/8] & (1<<(KEY_3 % 8))));
	printf("[vol-] %d\n", !!(key_b[KEY_4/8] & (1<<(KEY_4 % 8))));
}
	
    if (event.type == EV_KEY) {
      printf("==========================\n");
      printf("Code  : %02Xh (%s)\n", event.code, 
              ((name = key_name_from_code(event.code)) != NULL)? name: "none");
      printf("Value : %s\n", (event.value == DEMO_KEY_VALUE_DOWN)? "Down": "Up");
	  printf("--------------------------\n");
    }
  }
  
  return 0;
}
